@@ -18,37 +18,45 @@ from posixpath import normpath
 import six
 from six.moves import urllib
 from six.moves.urllib.parse import quote, unquote
+try:
+    from icecream import ic
+except ImportError:  # Graceful fallback if IceCream isn't installed.
+    ic = lambda *a: None if not a else (a[0] if len(a) == 1 else a)  # noqa
 
 from .omdict1D import omdict1D
 from .compat import string_types, UnicodeMixin
 from .common import (
     callable_attr, is_iterable_but_not_string, absent as _absent)
 
-#
-# TODO(grun): Subclass Path, PathCompositionInterface, Query, and
-# QueryCompositionInterface into two subclasses each - one for the URL
-# and one for the Fragment.
-#
-# Subclasses will clean up the code because the valid encodings are
-# different between a URL Path and a Fragment Path and a URL Query and a
-# Fragment Query.
-#
-# For example, '?' and '#' don't need to be encoded in Fragment Path
-# segments but must be encoded in URL Path segments.
-#
-# Similarly, '#' doesn't need to be encoded in Fragment Query keys and
-# values, but must be encoded in URL Query keys and values.
-#
 
-# Map of various URL schemes to their default ports. Scheme strings are
+# Map of common protocols, as suggested by the common protocols included in
+# urllib/parse.py, to their default ports. Protocol scheme strings are
 # lowercase.
 DEFAULT_PORTS = {
+    'ws': 80,
     'ftp': 21,
+    'git': 9418,
+    'hdl': 2641,
+    'nfs': 111,
+    'sip': 5060,
     'ssh': 22,
+    'svn': 3690,
+    'wss': 443,
     'http': 80,
+    'imap': 143,
+    'nntp': 119,
+    'sftp': 22,
+    'sips': 5061,
     'tftp': 69,
+    'rtsp': 554,
+    'wais': 210,
     'https': 443,
+    'rsync': 873,
+    'rtspu': 5004,
+    'snews': 563,
+    'gopher': 70,
     'telnet': 23,
+    'prospero': 191,
 }
 
 
@@ -217,6 +225,12 @@ def set_scheme(url, scheme):
         return '%s:%s' % (scheme, after_scheme)
 
 
+# 'netloc' in Python parlance, 'authority' in RFC 3986 parlance.
+def has_netloc(url):
+    scheme = get_scheme(url)
+    return url.startswith('//' if scheme is None else scheme + '://')
+
+
 def urlsplit(url):
     """
     Parameters:
@@ -256,6 +270,7 @@ def urlsplit(url):
         netloc = None
 
     scheme = original_scheme
+
     return urllib.parse.SplitResult(scheme, netloc, path, query, fragment)
 
 
@@ -267,16 +282,30 @@ def urljoin(base, url):
 
     Returns: The resultant URL from joining <base> and <url>.
     """
-    if not base:
-        return url
+    base_scheme = get_scheme(base) if has_netloc(base) else None
+    url_scheme = get_scheme(url) if has_netloc(url) else None
 
-    base_scheme = get_scheme(base)
-    url_scheme = get_scheme(url)
+    if base_scheme is not None:
+        # For consistent URL joining, switch the base URL's scheme to
+        # 'http'. urllib.parse.urljoin() behaves differently depending on the
+        # scheme. E.g.
+        #
+        #   >>> urllib.parse.urljoin('http://google.com/', 'hi')
+        #   'http://google.com/hi'
+        #
+        # vs
+        #
+        #   >>> urllib.parse.urljoin('asdf://google.com/', 'hi')
+        #   'hi'
+        root = set_scheme(base, 'http')
+    else:
+        root = base
 
-    http_base = set_scheme(base, 'http')
-    joined = urllib.parse.urljoin(http_base, url)
-    if not url_scheme:
-        joined = set_scheme(joined, base_scheme)
+    joined = urllib.parse.urljoin(root, url)
+
+    new_scheme = url_scheme if url_scheme is not None else base_scheme
+    if new_scheme is not None and has_netloc(joined):
+        joined = set_scheme(joined, new_scheme)
 
     return joined
 
@@ -300,6 +329,7 @@ def join_path_segments(*args):
     Returns: A list containing the joined path segments.
     """
     finals = []
+
     for segments in args:
         if not segments or segments == ['']:
             continue
@@ -314,6 +344,7 @@ def join_path_segments(*args):
             elif finals[-1] != '' and segments[0] == '' and len(segments) > 1:
                 segments = segments[1:]
             finals.extend(segments)
+
     return finals
 
 
@@ -360,6 +391,12 @@ def remove_path_segments(segments, remove):
     return ret
 
 
+def quacks_like_a_path_with_segments(obj):
+    return (
+        hasattr(obj, 'segments') and
+        is_iterable_but_not_string(obj.segments))
+
+
 class Path(object):
 
     """
@@ -413,13 +450,15 @@ class Path(object):
 
     def load(self, path):
         """
-        Load <path>, replacing any existing path. <path> can either be a
-        list of segments or a path string to adopt.
+        Load <path>, replacing any existing path. <path> can either be
+        a Path instance, a list of segments, a path string to adopt.
 
         Returns: <self>.
         """
         if not path:
             segments = []
+        elif quacks_like_a_path_with_segments(path):  # Path interface.
+            segments = path.segments
         elif is_iterable_but_not_string(path):  # List interface.
             segments = path
         else:  # String interface.
@@ -439,12 +478,14 @@ class Path(object):
 
     def add(self, path):
         """
-        Add <path> to the existing path. <path> can either be a list of
-        segments or a path string to append to the existing path.
+        Add <path> to the existing path. <path> can either be a Path instance,
+        a list of segments, or a path string to append to the existing path.
 
         Returns: <self>.
         """
-        if is_iterable_but_not_string(path):  # List interface.
+        if quacks_like_a_path_with_segments(path):  # Path interface.
+            newsegments = path.segments
+        elif is_iterable_but_not_string(path):  # List interface.
             newsegments = path
         else:  # String interface.
             newsegments = self._segments_from_path(path)
@@ -459,6 +500,7 @@ class Path(object):
             segments.insert(0, '')
 
         self.load(join_path_segments(segments, newsegments))
+
         return self
 
     def set(self, path):
@@ -475,6 +517,7 @@ class Path(object):
                 segments = self._segments_from_path(path)
             base = ([''] if self.isabsolute else []) + self.segments
             self.load(remove_path_segments(base, segments))
+
         return self
 
     def normalize(self):
@@ -488,6 +531,7 @@ class Path(object):
             if normalized.startswith('//'):  # http://bugs.python.org/636648
                 normalized = '/' + normalized.lstrip('/')
             self.load(normalized)
+
         return self
 
     def asdict(self):
@@ -537,7 +581,11 @@ class Path(object):
         return not self.isdir
 
     def __truediv__(self, path):
-        return self.add(path)
+        copy = self.__class__(
+            path=self.segments,
+            force_absolute=self._force_absolute,
+            strict=self.strict)
+        return copy.add(path)
 
     def __eq__(self, other):
         return str(self) == str(other)
@@ -855,6 +903,7 @@ class Query(object):
             else:
                 key = item
                 self.params.pop(key, None)
+
         return self
 
     @property
@@ -1115,6 +1164,7 @@ class Fragment(FragmentPathCompositionInterface, QueryCompositionInterface):
             self.path.add(path)
         if args is not _absent:
             self.query.add(args)
+
         return self
 
     def set(self, path=_absent, args=_absent, separator=_absent):
@@ -1124,6 +1174,7 @@ class Fragment(FragmentPathCompositionInterface, QueryCompositionInterface):
             self.query.load(args)
         if separator is True or separator is False:
             self.separator = separator
+
         return self
 
     def remove(self, fragment=_absent, path=_absent, args=_absent):
@@ -1133,6 +1184,7 @@ class Fragment(FragmentPathCompositionInterface, QueryCompositionInterface):
             self.path.remove(path)
         if args is not _absent:
             self.query.remove(args)
+
         return self
 
     def asdict(self):
@@ -1168,9 +1220,9 @@ class Fragment(FragmentPathCompositionInterface, QueryCompositionInterface):
         if path and (not query or not self.separator):
             path = path.replace('%3F', '?')
 
-        if query and path:
-            return path + ('?' if self.separator else '') + query
-        return path + query
+        separator = '?' if path and query and self.separator else ''
+
+        return path + separator + query
 
     def __repr__(self):
         return "%s('%s')" % (self.__class__.__name__, str(self))
@@ -1241,7 +1293,12 @@ class furl(URLPathCompositionInterface, QueryCompositionInterface,
       fragment: Fragment object from FragmentCompositionInterface.
     """
 
-    def __init__(self, url='', strict=False):
+    def __init__(self, url='', args=_absent, path=_absent, fragment=_absent,
+                 scheme=_absent, netloc=_absent, origin=_absent,
+                 fragment_path=_absent, fragment_args=_absent,
+                 fragment_separator=_absent, host=_absent, port=_absent,
+                 query=_absent, query_params=_absent, username=_absent,
+                 password=_absent, strict=False):
         """
         Raises: ValueError on invalid url.
         """
@@ -1251,6 +1308,12 @@ class furl(URLPathCompositionInterface, QueryCompositionInterface,
         self.strict = strict
 
         self.load(url)  # Raises ValueError on invalid url.
+        self.set(
+            args=args, path=path, fragment=fragment, scheme=scheme,
+            netloc=netloc, origin=origin, fragment_path=fragment_path,
+            fragment_args=fragment_args, fragment_separator=fragment_separator,
+            host=host, port=port, query=query, query_params=query_params,
+            username=username, password=password)
 
     def load(self, url):
         """
@@ -1268,8 +1331,7 @@ class furl(URLPathCompositionInterface, QueryCompositionInterface,
             url = str(url)
 
         # urlsplit() raises a ValueError on malformed IPv6 addresses in
-        # Python 2.7+. In Python <= 2.6, urlsplit() doesn't raise a
-        # ValueError on malformed IPv6 addresses.
+        # Python 2.7+.
         tokens = urlsplit(url)
 
         self.netloc = tokens.netloc  # Raises ValueError in Python 2.7+.
@@ -1279,6 +1341,7 @@ class furl(URLPathCompositionInterface, QueryCompositionInterface,
         self.path.load(tokens.path)
         self.query.load(tokens.query)
         self.fragment.load(tokens.fragment)
+
         return self
 
     @property
@@ -1410,6 +1473,7 @@ class furl(URLPathCompositionInterface, QueryCompositionInterface,
         if self.port and self.port != DEFAULT_PORTS.get(self.scheme):
             port = ':%s' % self.port
         origin = '%s://%s%s' % (scheme, host, port)
+
         return origin
 
     @origin.setter
@@ -1479,6 +1543,7 @@ class furl(URLPathCompositionInterface, QueryCompositionInterface,
             self.query.add(query_params)
         if fragment_path is not _absent or fragment_args is not _absent:
             self.fragment.add(path=fragment_path, args=fragment_args)
+
         return self
 
     def set(self, args=_absent, path=_absent, fragment=_absent, scheme=_absent,
@@ -1718,7 +1783,7 @@ class furl(URLPathCompositionInterface, QueryCompositionInterface,
             }
 
     def __truediv__(self, path):
-        return self.add(path=path)
+        return self.copy().add(path=path)
 
     def __eq__(self, other):
         try:
